@@ -34,13 +34,9 @@ def fitForecast(y, h, sumMat, nodes, method, freq, include_history, cap, capF, c
     # If you have a ditionary of Prophet Dataframes already, skip the prophet part, and put all the values into a dictionary
     ##
     if skipFitting == True:
-        i = -2
-        for column in y:
-            i += 1
-            if i == -1:
-                continue
-            else:
-                forecastsDict[i] = y[column]
+        for key in range(len(y.columns.tolist())-1):
+            forecastsDict[key] = pd.DataFrame(y.iloc[:,key+1], columns = ["yhat"])
+            forecastsDict[key] = forecastsDict[key].rename(columns = {forecastsDict[key].columns[1] : 'yhat'})
             
     if skipFitting == False:
         
