@@ -21,16 +21,17 @@ from sklearn.model_selection import TimeSeriesSplit
 from htsprophet.fitForecast import fitForecast
 from scipy.stats import boxcox
 from scipy.special import inv_boxcox
+from multiprocessing.dummy import Pool as ThreadPool
 
 
 #%%
 def hts(y, h = 1, nodes = [[2]], method='OLS', freq = 'D', transform = None, include_history = True, cap = None, capF = None, changepoints = None, \
         n_changepoints = 25, yearly_seasonality = 'auto', weekly_seasonality = 'auto', holidays = None, seasonality_prior_scale = 10.0, \
-        holidays_prior_scale = 10.0, changepoint_prior_scale = 0.05, mcmc_samples = 0, interval_width = 0.80, uncertainty_samples = 0, skipFitting = False):
+        holidays_prior_scale = 10.0, changepoint_prior_scale = 0.05, mcmc_samples = 0, interval_width = 0.80, uncertainty_samples = 0, skipFitting = False, numThreads = 0):
     '''
     Parameters
     ----------------
-     y - dataframe of time-series data
+     y - dataframe of time-series data, or if you want to skip fitting, a dictionary of prophet base forecast dataframes
                Layout:
                    0th Col - Time instances
                    1st Col - Total of TS
@@ -82,6 +83,10 @@ def hts(y, h = 1, nodes = [[2]], method='OLS', freq = 'D', transform = None, inc
      n_changepoints - (constant or list) changepoints for the model to consider fitting. If it is a list, then
                                          the number of items must equal len(y.columns) - 1
                                          
+     skipFitting - (Boolean) if y is already a dictionary of dataframes, set this to True, and DO NOT run with method = "cvSelect" or transform = "BoxCox"
+     
+     numThreads - (int) number of threads you want to use when running cvSelect. Note: 14 has shown to decrease runtime by 10 percent 
+                                 
      All other inputs - see Prophet
      
     Returns
@@ -130,25 +135,26 @@ def hts(y, h = 1, nodes = [[2]], method='OLS', freq = 'D', transform = None, inc
         return finalMat
     ##
     # Error Handling
-    ##
+    ##        
     if h < 1:
         sys.exit('you must set h (number of step-ahead forecasts) to a positive number')
     if method not in ['OLS','WLSS','WLSV','FP','PHA','AHP','BU','cvSelect']:
         sys.exit("not a valid method input, must be one of the following: 'OLS','WLSS','WLSV','FP','PHA','AHP','BU','cvSelect'")
     if len(nodes) < 1:
         sys.exit("nodes input should at least be of length 1")
-    if sum(list(map(sum, nodes))) != len(y.columns)-2:
-        sys.exit("The sum of the nodes list does not equal the number of columns - 2, dataframe should contain a time column in the 0th pos. Double check node input")
     if not isinstance(cap, int) and not isinstance(cap, pd.DataFrame) and not isinstance(cap, float) and not cap is None:
         sys.exit("cap should be a constant (float or int) or a DataFrame, or not specified")
     if not isinstance(capF, int) and not isinstance(capF, pd.DataFrame) and not isinstance(capF, float) and not capF is None:
         sys.exit("capF should be a constant (float or int) or a DataFrame, or not specified")
-    if isinstance(cap, pd.DataFrame):
-        if len(cap.columns) != len(y.columns)-1:
-            sys.exit("If cap is a DataFrame, it should have a number of columns equal to the input Dataframe - 1")
-    if isinstance(capF, pd.DataFrame):
-        if len(capF.columns) != len(y.columns)-1:
-            sys.exit("If capF is a DataFrame, it should have a number of columns equal to the input Dataframe - 1")
+    if not isinstance(y, dict):
+        if sum(list(map(sum, nodes))) != len(y.columns)-2:
+            sys.exit("The sum of the nodes list does not equal the number of columns - 2, dataframe should contain a time column in the 0th pos. Double check node input")
+        if isinstance(cap, pd.DataFrame):
+            if len(cap.columns) != len(y.columns)-1:
+                sys.exit("If cap is a DataFrame, it should have a number of columns equal to the input Dataframe - 1")
+        if isinstance(capF, pd.DataFrame):
+            if len(capF.columns) != len(y.columns)-1:
+                sys.exit("If capF is a DataFrame, it should have a number of columns equal to the input Dataframe - 1")
     if cap is not None and method not in ["BU","FP","AHP","PHA"]:
         print("Consider using BU, FP, AHP, or PHA.  The other methods can create negatives which would cause problems for the log() function")
     ##
@@ -200,36 +206,45 @@ def hts(y, h = 1, nodes = [[2]], method='OLS', freq = 'D', transform = None, inc
                     y1.iloc[:,column+1] = inv_boxcox(y1.iloc[:, column+1], boxcoxT[column])
                     
         for trainIndex, testIndex in tscv.split(y.iloc[:,0]):
-            ynew1 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[0], freq, include_history, cap, capF, changepoints, n_changepoints, \
-                                yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
-                                changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
-            ynew2 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[1], freq, include_history, cap, capF, changepoints, n_changepoints, \
-                                yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
-                                changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
-            ynew3 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[2], freq, include_history, cap, capF, changepoints, n_changepoints, \
-                                yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
-                                changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
-            ynew4 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[3], freq, include_history, cap, capF, changepoints, n_changepoints, \
-                                yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
-                                changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
-            ynew5 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[4], freq, include_history, cap, capF, changepoints, n_changepoints, \
-                                yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
-                                changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
-            ynew6 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[5], freq, include_history, cap, capF, changepoints, n_changepoints, \
-                                yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
-                                changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
-            ynew7 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[6], freq, include_history, cap, capF, changepoints, n_changepoints, \
-                                yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
-                                changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
-                    
+            if numThreads != 0:
+                pool = ThreadPool(numThreads)
+                results = pool.starmap(fitForecast, zip([y.iloc[trainIndex, :]]*7, [len(testIndex)]*7, [sumMat]*7, [nodes]*7, methodList, [freq]*7, [include_history]*7, [cap]*7, [capF]*7, [changepoints]*7, [n_changepoints]*7, \
+                                    [yearly_seasonality]*7, [weekly_seasonality]*7, [holidays]*7, [seasonality_prior_scale]*7, [holidays_prior_scale]*7,\
+                                    [changepoint_prior_scale]*7, [mcmc_samples]*7, [interval_width]*7, [uncertainty_samples]*7,  [boxcoxT]*7, [skipFitting]*7))
+                pool.close()
+                pool.join()
+                ynew1, ynew2, ynew3, ynew4, ynew5, ynew6, ynew7 = results
+            else:
+                ynew1 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[0], freq, include_history, cap, capF, changepoints, n_changepoints, \
+                                    yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
+                                    changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
+                ynew2 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[1], freq, include_history, cap, capF, changepoints, n_changepoints, \
+                                    yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
+                                    changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
+                ynew3 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[2], freq, include_history, cap, capF, changepoints, n_changepoints, \
+                                    yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
+                                    changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
+                ynew4 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[3], freq, include_history, cap, capF, changepoints, n_changepoints, \
+                                    yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
+                                    changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
+                ynew5 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[4], freq, include_history, cap, capF, changepoints, n_changepoints, \
+                                    yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
+                                    changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
+                ynew6 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[5], freq, include_history, cap, capF, changepoints, n_changepoints, \
+                                    yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
+                                    changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
+                ynew7 = fitForecast(y.iloc[trainIndex, :], len(testIndex), sumMat, nodes, methodList[6], freq, include_history, cap, capF, changepoints, n_changepoints, \
+                                    yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
+                                    changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
+#                    
             for key in ynew1.keys():
-                MASE1.append(abs(ynew1[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)/np.mean(abs(y1.iloc[trainIndex[1:], key + 1].values - y1.iloc[trainIndex[:-1], key + 1].values)))
-                MASE2.append(abs(ynew2[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)/np.mean(abs(y1.iloc[trainIndex[1:], key + 1].values - y1.iloc[trainIndex[:-1], key + 1].values)))
-                MASE3.append(abs(ynew3[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)/np.mean(abs(y1.iloc[trainIndex[1:], key + 1].values - y1.iloc[trainIndex[:-1], key + 1].values)))
-                MASE4.append(abs(ynew4[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)/np.mean(abs(y1.iloc[trainIndex[1:], key + 1].values - y1.iloc[trainIndex[:-1], key + 1].values)))
-                MASE5.append(abs(ynew5[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)/np.mean(abs(y1.iloc[trainIndex[1:], key + 1].values - y1.iloc[trainIndex[:-1], key + 1].values)))
-                MASE6.append(abs(ynew6[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)/np.mean(abs(y1.iloc[trainIndex[1:], key + 1].values - y1.iloc[trainIndex[:-1], key + 1].values)))
-                MASE7.append(abs(ynew7[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)/np.mean(abs(y1.iloc[trainIndex[1:], key + 1].values - y1.iloc[trainIndex[:-1], key + 1].values)))
+                MASE1.append(np.mean(abs(ynew1[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)))
+                MASE2.append(np.mean(abs(ynew2[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)))
+                MASE3.append(np.mean(abs(ynew3[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)))
+                MASE4.append(np.mean(abs(ynew4[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)))
+                MASE5.append(np.mean(abs(ynew5[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)))
+                MASE6.append(np.mean(abs(ynew6[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)))
+                MASE7.append(np.mean(abs(ynew7[key].yhat[-len(testIndex):].values - y1.iloc[testIndex, key+1].values)))
         ##
         # If the method has the minimum Average MASE, use it on all of the data
         ##
@@ -241,16 +256,35 @@ def hts(y, h = 1, nodes = [[2]], method='OLS', freq = 'D', transform = None, inc
         print(choice)
     
     else:    
+        if skipFitting == True:
+            theDictionary = y
+            i = 0
+            for key in y.keys():
+                if i == 0:
+                    y = pd.DataFrame(theDictionary[key].ds)
+                y[i] = theDictionary[key].yhat
+                i += 1
         sumMat = SummingMat(nodes)
         ynew = fitForecast(y, h, sumMat, nodes, method, freq, include_history, cap, capF, changepoints, n_changepoints, \
                            yearly_seasonality, weekly_seasonality, holidays, seasonality_prior_scale, holidays_prior_scale,\
                            changepoint_prior_scale, mcmc_samples, interval_width, uncertainty_samples, boxcoxT, skipFitting)
-        
+    ##
+    # Inverse boxcox the data
+    ##    
     if transform is not None:
         if transform == 'BoxCox':
             for column in range(len(y.columns.tolist())-1):
                 y.iloc[:,column+1] = inv_boxcox(y.iloc[:, column+1], boxcoxT[column])
     ##
+    # Put the values back in the dictionary for skipFitting
+    ##
+    if skipFitting == True:
+        i = 0
+        for key in theDictionary.keys():
+            theDictionary[key].yhat = ynew[i].yhat
+            i += 1
+            ynew = theDictionary
+    ##
     # Rename keys so that dictionary can be easily understood
     ##
     i = -2
