@@ -18,8 +18,8 @@ It was my intention to make some of the code look similar to certain sections in
 """
 
 import pandas as pd
-#from htsprophet.hts import hts, orderHier, makeWeekly
-#from htsprophet.htsPlot import plotNode, plotChild, plotNodeComponents
+from htsprophet.hts import hts, orderHier, makeWeekly
+from htsprophet.htsPlot import plotNode, plotChild, plotNodeComponents
 import numpy as np
 
 #%% Random data (Change this to whatever data you want)
@@ -38,7 +38,7 @@ data["businessMarket"] = busDat
 data["sessions"] = sessions
 data1 = makeWeekly(data)
 data2, nodes = orderHier(data1, 1, 2, 3)
-data2.iloc[:,1:] = np.log(data2.iloc[:,1:]+1)
+#data2.iloc[:,1:] = np.log(data2.iloc[:,1:]+1)
 #%% Run HTS
 ##
 # Make the daily data weekly (optional)
@@ -52,11 +52,11 @@ data2.iloc[:,1:] = np.log(data2.iloc[:,1:]+1)
 ## load in prophet inputs (Running HTS runs prophet, so all inputs should be gathered beforehand)
 ## Made up holiday data
 ###
-#holidates = pd.date_range("12/25/2013","12/31/2017", freq = 'A')
-#holidays = pd.DataFrame(["Christmas"]*5, columns = ["holiday"])
-#holidays["ds"] = holidates
-#holidays["lower_window"] = [-4]*5
-#holidays["upper_window"] = [0]*5
+holidates = pd.date_range("12/25/2013","12/31/2017", freq = 'A')
+holidays = pd.DataFrame(["Christmas"]*5, columns = ["holiday"])
+holidays["ds"] = holidates
+holidays["lower_window"] = [-4]*5
+holidays["upper_window"] = [0]*5
 ##
 # Run hts with the CVselect function (this decides which hierarchical aggregation method to use based on minimum mean Mean Absolute Scaled Error)
 # h (which is 12 here) - how many steps ahead you would like to forecast.  If youre using daily data you don't have to specify freq.
@@ -64,11 +64,13 @@ data2.iloc[:,1:] = np.log(data2.iloc[:,1:]+1)
 # NOTE: CVselect takes a while, so if you want results in minutes instead of half-hours pick a different method
 ##
 #%%
-myDict = hts(data2, 52, nodes, method = "y")
+import random
+random.seed(0)
+myDict = hts(data2, 52, nodes, method = "cvSelect", holidays = holidays, numThreads = 4)
 #
 # This output is a dictionary of dataframes, so you can do any further analysis that you may want. It also allows you to plot the forecasts.
 # Some functions I've made are: (1 means I'm plotting the total node)
 ##
-#plotNode(myDict, "Total", h = 52, xlabel = "Week", ylabel = "Number of Sessions")
-#plotChild(myDict, "Total", h = 52, xlabel = "Week", ylabel = "Number of Sessions")
-#plotNodeComponents(myDict, "Total", holidays = holidays)
\ No newline at end of file
+plotNode(myDict, "Total", h = 52, xlabel = "Week", ylabel = "Number of Sessions")
+plotChild(myDict, "Total", h = 52, xlabel = "Week", ylabel = "Number of Sessions")
+plotNodeComponents(myDict, "Total", holidays = holidays)
\ No newline at end of file
